#!/usr/bin/env python

import time

# Import OpenMM tools
from simtk import openmm, unit
from simtk.openmm.app import *

# Use MDTraj to write simulation trajectories
from mdtraj.reporters import NetCDFReporter

# Import the SMIRNOFF force field engine and some useful tools
from openff.toolkit.typing.engines.smirnoff import ForceField
# LPW: Open Force Field's PME is different from openmm's PME
from openff.toolkit.typing.engines.smirnoff.forcefield import PME
from openff.toolkit.utils import get_data_file_path

# Import the OpenEye toolkit
from openeye import oechem

pdb_filename = 'ETH-box.pdb'
mol_filename = 'ETH.mol2'

# Define a few simulation parameters
time_step = 2*unit.femtoseconds # simulation timestep
temperature = 300*unit.kelvin # simulation temperature
friction = 1/unit.picosecond # collision rate
num_steps = 1000000 # number of steps to run
trj_freq = 1000 # number of steps per written trajectory frame
data_freq = 1000 # number of steps per written simulation statistics

# Load molecule and create pdb object
pdb = PDBFile(pdb_filename)

# Load a SMIRNOFF force field
forcefield = ForceField(get_data_file_path('test_forcefields/Frosst_AlkEthOH_parmAtFrosst.offxml'))

# Load molecule using OpenEye tools
mol = oechem.OEGraphMol()
ifs = oechem.oemolistream(mol_filename)
# LPW: I don't understand the meaning of these lines.
# flavor = oechem.OEIFlavor_Generic_Default | oechem.OEIFlavor_MOL2_Default | oechem.OEIFlavor_MOL2_Forcefield
# ifs.SetFlavor( oechem.OEFormat_MOL2, flavor)
oechem.OEReadMolecule(ifs, mol)
oechem.OETriposAtomNames(mol)

# Create the OpenMM system
system = forcefield.createSystem(pdb.topology, [mol], nonbondedMethod=PME, nonbondedCutoff=1.0*unit.nanometers, rigidWater=True)

# Set up an OpenMM simulation
integrator = openmm.LangevinIntegrator(temperature, friction, time_step)
platform = openmm.Platform.getPlatformByName('CUDA')
simulation = Simulation(pdb.topology, system, integrator)
simulation.context.setPositions(pdb.positions)
simulation.context.setVelocitiesToTemperature(temperature)
netcdf_reporter = NetCDFReporter('water_traj.nc', trj_freq)
simulation.reporters.append(netcdf_reporter)
simulation.reporters.append(StateDataReporter('water_data.csv', data_freq, step=True, potentialEnergy=True, temperature= True, density=True))

print(simulation.context.getState(getEnergy=True).getPotentialEnergy())
simulation.minimizeEnergy()
print(simulation.context.getState(getEnergy=True).getPotentialEnergy())

# Run the simulation
print("Starting simulation")
start = time.clock()
simulation.step(num_steps)
end = time.clock()

print(f"Elapsed time {end-start:.2f} seconds")
netcdf_reporter.close()
print("Done!")



import pathlib
import re
from urllib.request import Request, urlopen

import pytest


def find_readme_links() -> list[str]:
    """Yield all the links in the main README.md file.

    Returns
    -------
    readme_examples : list[str]
        The list of links included in the README.md file.
    """
    if "site-packages" in __file__:
        # This test file is being collected from the installed package, which
        # does not provide the README file.
        # Note that there will likely be a mis-bundled file
        # $CONDA_PREFIX/lib/python3.x/site-packages/README.md, but this is not
        # the toolkit's README file!
        return list()

    else:
        readme_file_path = pathlib.Path(__file__).parents[3] / "README.md"
        with open(readme_file_path.as_posix()) as f:
            readme_content = f.read()

    with open(readme_file_path.as_posix()) as f:
        readme_content = f.read()

    return re.findall("http[s]?://(?:[0-9a-zA-Z]|[-/.%:_])+", readme_content)


@pytest.mark.parametrize("readme_link", find_readme_links())
def test_readme_links(readme_link):
    """Test URLs in README are reachable"""
    # Some websites do not accept requests that don't specify the
    # client and the type of accepted documents so we add fake info
    # to avoid the response being an error.
    headers = {
        "User-Agent": "Mozilla/5.0",
        "Accept": "application/xhtml+xml,text/html,application/xml;q=0.9,*/*;q=0.8",
    }
    request = Request(readme_link, headers=headers)

    # Some DOI-based links are now behind DDoS protections, so skip them
    if "doi.org" in readme_link:
        pytest.skip("DOI links are behind DDoS protection and do not resolve")
    if "codecov.io" in readme_link:
        pytest.skip("Codecov website DOI also may be behind DDoS protection")
    if "export" in readme_link:
        # October 2023 Zenodo upgrade seems to have broken this link - although
        # it is still accessible from the web UI (go to the landing page of a
        # record, Export section (select BibTeX in drop-down) and click "Export"
        # button will bring you to this link.
        pytest.skip("Seems to be a Zenodo regression in bibtex link")
    if readme_link.endswith("MIT"):
        # January 22 2024: Link causing some failures, accessible in browser
        pytest.skip("Flaky")

    # Try to connect 5 times, keeping track of exceptions so useful feedback can be provided.
    success = False
    exception = None
    for retry in range(5):
        try:
            urlopen(request)
            success = True
            break
        except Exception as e:
            exception = e
    if not (success):
        raise exception



"""
openff-test-parameter-plugins
A test package used to ensure that parameterhandler plugins are handled correctly
"""

from setuptools import setup

setup(
    name="openff-test-parameter-plugins",
    packages=["custom_plugins"],
    include_package_data=True,
    entry_points={
        "openff.toolkit.plugins.handlers": [
            "CustomHandler = custom_plugins.handler_plugins:CustomHandler",
            "WrongSubclass = custom_plugins.handler_plugins:WrongSubclass",
            "FOOBuckinghamHandler = custom_plugins.handler_plugins:FOOBuckinghamHandler",
        ]
    },
)



import os

from _cif_to_substructure_dict import CifSubstructures

from openff.toolkit.utils import get_data_file_path

if not os.path.exists("aa-variants-v1.cif"):
    import requests

    r = requests.get("https://ftp.wwpdb.org/pub/pdb/data/monomers/aa-variants-v1.cif")
    print(r.ok)
    if r.ok:
        with open("aa-variants-v1.cif", "wb") as of:
            for chunk in r.iter_content(chunk_size=1024 * 8):
                if chunk:
                    of.write(chunk)
                    of.flush()
                    os.fsync(of.fileno())


cif_object = CifSubstructures()
cif_object.from_file(
    "aa-variants-v1.cif",
    replace_quadruple_bond_with_any=True,
    remove_charge_bond_order_resonant=True,
)

# Automatically add known substructures that are missing from aa_variants but expected by force fields (like ACE and
# NME caps)
cif_object._add_common_substructures()

output_file = get_data_file_path("proteins/T4-protein.sdf").replace(
    "T4-protein.sdf", "aa_residues_substructures_with_caps.json"
)
cif_object.to_json_file(output_file)



"""
XML I/O parser for the SMIRNOFF (SMIRKS Native Open Force Field) format.

.. codeauthor:: John D. Chodera <john.chodera@choderalab.org>
.. codeauthor:: David L. Mobley <dmobley@mobleylab.org>
.. codeauthor:: Peter K. Eastman <peastman@stanford.edu>

"""

__all__ = [
    "ParameterIOHandler",
    "XMLParameterIOHandler",
]
import logging
from collections.abc import Iterable

import xmltodict

logger = logging.getLogger(__name__)


class ParameterIOHandler:
    """
    Base class for handling serialization/deserialization of SMIRNOFF ForceField objects
    """

    _FORMAT: str

    def __init__(self):
        """
        Create a new ParameterIOHandler.

        """
        pass

    def parse_file(self, file_path):
        """

        Parameters
        ----------
        file_path

        Returns
        -------

        """
        pass

    def parse_string(self, data):
        """
        Parse a SMIRNOFF force field definition in a seriaized format

        Parameters
        ----------
        data

        Returns
        -------

        """
        pass

    def to_file(self, file_path, smirnoff_data):
        """
        Write the current force field parameter set to a file.

        Parameters
        ----------
        file_path
            The path to the file to write to.
        smirnoff_data
            A dictionary structured in compliance with the SMIRNOFF spec

        Returns
        -------

        """
        pass

    def to_string(self, smirnoff_data):
        """
        Render the force field parameter set to a string

        Parameters
        ----------
        smirnoff_data
            A dictionary structured in compliance with the SMIRNOFF spec

        Returns
        -------
        str
        """
        pass


class XMLParameterIOHandler(ParameterIOHandler):
    """
    Handles serialization/deserialization of SMIRNOFF ForceField objects from OFFXML format.
    """

    # TODO: Come up with a better keyword for format
    _FORMAT = "XML"

    def parse_file(self, source):
        """Parse a SMIRNOFF force field definition in XML format, read from a file.

        Parameters
        ----------
        source
            File path of file-like object implementing a ``read()`` method
            specifying a SMIRNOFF force field definition in `the SMIRNOFF XML format
            <https://openforcefield.github.io/standards/standards/smirnoff/#xml-representation>`_.

        Raises
        ------
        SMIRNOFFParseError
            If the XML cannot be processed.
        FileNotFoundError
            If the file could not found.

        """
        # If this is a file-like object, we should be able to read it.
        try:
            raw_data = source.read()
        except AttributeError:
            # This raises FileNotFoundError if the file doesn't exist.
            with open(source) as source_obj:
                raw_data = source_obj.read()

        # Parse the data in string format.
        return self.parse_string(raw_data)

    def parse_string(self, data: str) -> dict:
        """Parse a SMIRNOFF force field definition in XML format.

        A ``SMIRNOFFParseError`` is raised if the XML cannot be processed.

        Parameters
        ----------
        data
            A SMIRNOFF force field definition in `the SMIRNOFF XML format
            <https://openforcefield.github.io/standards/standards/smirnoff/#xml-representation>`_.

        """
        from pyexpat import ExpatError

        from openff.toolkit.utils.exceptions import SMIRNOFFParseError

        # Parse XML file
        try:
            smirnoff_data = xmltodict.parse(data, attr_prefix="")
            return smirnoff_data
        except ExpatError as e:
            raise SMIRNOFFParseError(str(e))

    def to_file(self, file_path, smirnoff_data):
        """Write the current force field parameter set to a file.

        Parameters
        ----------
        file_path
            The path to the file to be written.
            The `.offxml` or `.xml` file extension must be present.
        smirnoff_data
            A dict structured in compliance with the SMIRNOFF data spec.

        """
        xml_string = self.to_string(smirnoff_data)
        with open(file_path, "w") as of:
            of.write(xml_string)

    def to_string(self, smirnoff_data: dict) -> str:
        """
        Write the current force field parameter set to an XML string.

        Parameters
        ----------
        smirnoff_data
            A dictionary structured in compliance with the SMIRNOFF spec

        Returns
        -------
        serialized_forcefield
            XML String representation of this force field.

        """

        def prepend_all_keys(
            d: dict,
            char: str | None = "@",
            ignore_keys: Iterable[str] = frozenset(),
        ):
            """
            Modify a dictionary in-place, prepending a specified string to each key
            that doesn't refer to a value that is list or dict.

            Parameters
            ----------
            d
                Hierarchical dictionary to traverse and modify keys
            char
                String to prepend onto each applicable dictionary key
            ignore_keys
                A set or list of strings, indicating keys not to prepend in the data structure

            """
            if isinstance(d, dict):
                for key in list(d.keys()):
                    if key in ignore_keys:
                        continue
                    if isinstance(d[key], list) or isinstance(d[key], dict):
                        prepend_all_keys(d[key], char=char, ignore_keys=ignore_keys)
                    else:
                        new_key = char + key
                        d[new_key] = d[key]
                        del d[key]
                        prepend_all_keys(d[new_key], char=char, ignore_keys=ignore_keys)
            elif isinstance(d, list):
                for item in d:
                    prepend_all_keys(item, char=char, ignore_keys=ignore_keys)

        # the "xmltodict" library defaults to print out all element attributes on separate lines
        # unless they're prepended by "@"
        prepend_all_keys(smirnoff_data["SMIRNOFF"], ignore_keys=["Author", "Date"])

        # Reorder parameter sections to put Author and Date at the top (this is the only
        # way to change the order of items in a dict, as far as I can tell)
        for key, value in list(smirnoff_data["SMIRNOFF"].items()):
            if key in ["Author", "Date"]:
                continue
            del smirnoff_data["SMIRNOFF"][key]
            smirnoff_data["SMIRNOFF"][key] = value

        return xmltodict.unparse(smirnoff_data, pretty=True, indent=" " * 4)
